#!/usr/bin/perl -w

# DGAS Pushd.
# written by R.M. Piro (piro@to.infn.it),
# A. Guarise (guarise@to.infn.it)
# and G. Patania (patania@to.infn.it), INFN Torino, Italy.

use strict;
use POSIX;
use Text::ParseWords;

# turn off buffering of STDOUT
$| = 1;

START:

#FIXME needs to understand the jobmanager type blah vs lcgpbs and set it up

my $sigset = POSIX::SigSet ->new();
my $actionHUP=POSIX::SigAction->new("sigHUP_handler",$sigset,&POSIX::SA_NODEFER);
my $actionInt=POSIX::SigAction->new("sigINT_handler",$sigset,&POSIX::SA_NODEFER);
POSIX::sigaction(&POSIX::SIGHUP, $actionHUP);
POSIX::sigaction(&POSIX::SIGINT, $actionInt);
POSIX::sigaction(&POSIX::SIGTERM, $actionInt);
my $configFilePath ="/opt/glite/etc/glite-dgas-ce-pushd.conf"; 
my %configValues = (
		    dgasURDir         =>  "/opt/glite/var/dgasURBox",
		    dgasErrDir        =>  "/opt/glite/var/dgasURBox/ERR/",
		    qDepth 	      => "10",
		    qMult	      => "3",
		    lockFileName      => "/opt/glite/var/glite-dgas-ce-pushd.lock",
		    mainPollInterval  => "5",
		    queuePollInterval =>"25",
		    defaultVOUserHLRlist => "",
		    defaultUserHLR    =>"",
		    # forceLocalFirst   =>"yes", OBSOLETE!
		    forceLocalOnly    =>"yes",
		    gridUser	      =>"",
		    siteName	      =>"",
		    localHostNameAsCEHostName => "no",
		    );

my %voDefUserHLRs = ();

my $hostName = `hostname -s`;
chomp($hostName);
my $domainName = `hostname -d`;
chomp($domainName);

#Parse configuration file
if(exists $ARGV[0]) 
{
    $configFilePath = $ARGV[0];
}	
&parseConf($configFilePath); 

## Global variable initialization
my %urGridInfo;
my %urAcctlogInfo;
my $UR;
my @queue = ();

my $qDept = $configValues{qDepth};
my $qMult = $configValues{qMult}; 

my $dgas_dir     = $configValues{dgasURDir};
my $dgas_err_dir = $configValues{dgasErrDir};

my $lockFileName = $configValues{lockFileName};

my $T2 = $configValues{queuePollInterval};
my $T1 = $configValues{mainPollInterval};

my $defaultUserHLR =$configValues{defaultUserHLR};

#my $forceLocalFirst =$configValues{forceLocalFirst}; OBSOLETE
my $forceLocalFirst = "yes"; # now always yes! first ResourceHLR, then UserHLR

# check wether we _have_ to use $forceLocalFirst:
my $dgU = 0;
if ( ($dgU=&checkDgasUser($configValues{gridUser})) != 0) {
    if ($dgU < 0) {
	print "No login specified for calling the atmClient (gridUser in conf file)! Trying to use pool accounts ...!\n";
    }
    if ($dgU > 0) {
	print "Warning: Login specified for calling the atmClient (gridUser in conf file) cannot be used (either 'root' or without access to a shell)!";
	$forceLocalFirst = "yes";
    }
}

if ($forceLocalFirst eq "yes") {
    print "Sending all usage records to the Resource HLR first!\n";
}

my $forceLocalOnly =$configValues{forceLocalOnly};
if ($forceLocalOnly eq "yes") {
    print "Sending all usage records to the Resource HLR _only_!\n";
}

my @dvec = split(/\//, $dgas_err_dir);

my $errDirName =  $dvec[$#dvec];
my $keepGoing = 1;
my $start = time();   ##  time init

my $vomsProxyInfo = "";
#put lock
if ( &putLock($lockFileName) != 0 )
{
    &error("Fatal Error: Couldn't open lock file! in $lockFileName\n");
}
else
{
	print "".localtime().": Lock file succesfully created.\n";
}

( -d $configValues{dgasURDir} ) || mkdir $configValues{dgasURDir};

#create dir of errors
&createErrDir();

if ( &findVomsProxyInfo($vomsProxyInfo) == 0 )
{
        print "Found voms-proxy-info: $vomsProxyInfo\n";
}
else
{
        print "Could not find voms-proxy-info\n";
}
#print "Notice: Found voms-proxy-info: $vomsProxyInfo\n";


while( $keepGoing )
{
    
##----------------------------------------------------------------  STEP1
    opendir(DIR, $dgas_dir) || &error("can't opendir $dgas_dir: $!");

    # first get all interesting files from dgasURBox:
    my @urFiles = ();
    while( defined(my $file = readdir(DIR)) && $keepGoing ) 
    {
	next if ($file =~ /^\.\.?$/ || $file =~ /.proxy$/ || $file =~ /^$errDirName$/);	# skip . .. proxy files
	push (@urFiles, $file);
    }
    closedir DIR;

    # process files:
    while ($keepGoing && @urFiles)
    {
	my $file = pop (@urFiles);
	#print "$file ...\n";

	next if (! -r "$dgas_dir/$file" && ! -f "$dgas_dir/$file");

	if( &checkQueue($file) == 0 )           ## if file is not in queue
	{  
	    &parseFile($file);         ## if file is a true input file   
	    #  &printOut;
	    if(&callAtmClient($file) == 0)
	    {
		print "UR successfully forwarded to HLR.\n";
		&delFile($file);
	    }
	    else{&pushToQueue($file);}
	}
    }

    if ( $keepGoing )
    {
	my $secsWaited = 0;
	while ($keepGoing && $secsWaited < $T1) {
	    sleep 1;
	    $secsWaited++;
	}
    }

##----------------------------------------------------------------  END STEP1 
    
    if((&timeOut() == 0) && ($#queue != -1) && $keepGoing )    ##timeout reached
    {
	my @localQueue = ();
	
	@queue = sort { ( ($a =~ /:(\d*):\d*$/)[0] <=> ($b =~ /:(\d*):\d*$/)[0] )
			    || 
			    ( ($a =~ /:(\d*)$/)[0] <=> ($b =~ /:(\d*)$/)[0] ) } @queue;
	
	foreach(@queue)
	{
	    my @vec = split(/:/, $_);
	    
	    
	    if(-e $dgas_dir.$vec[0])
	    {
		print "".localtime().": Retrying to process $vec[0]; multiplicity: $vec[1]; queue depth: $vec[2]\n";

		&parseFile($vec[0]);
		if(&callAtmClient($vec[0]) == 0)
		{
		    print "UR successfully forwarded to HLR.\n";
		    &delFile($vec[0]);
		}
		elsif ($vec[1]>= $qMult)
		{
		    print "Warning: UR could not be forwarded to HLR; maximum number of retries reached; moving $vec[0] to $dgas_err_dir\n";
		    &moveInErrDir($vec[0]);
		}
		else
		{
		    print "Warning: UR could not be forwarded to HLR; retrying later ...\n";
		    my $updatedItem = &setFile(@vec);
		    push @localQueue , $updatedItem;
		}
	    }
	}
	
	@queue =  @localQueue;  #  update queue 
	splice @localQueue, 0;
    }
    
}   

&Exit();

sub Exit ()
{
	print "".localtime().": Exiting...\n";
	if ( &delLock($lockFileName) != 0 )
	{	
	    print "Error removing lock file.\n";
	}
	else
	{
		print "Lock file removed.\n";
	}
	print "Exit.\n";
}

###-----------------------END-----------------------------------------###



sub parseConf 
{
    my $fconf = $_[0];
    open(FILE, "$fconf") || &error("Error: Cannot open configuration file: $fconf\n");
    while(<FILE>)
    {
	if (/\$\{(.*)\}/)
        {
                my $value=$ENV{$1};
                s/\$\{$1\}/$value/g;
        }
	if(/^qDept\s*=\s*\"(.*)\"$/){$configValues{qDept}=$1;}
	if(/^qMult\s*=\s*\"(.*)\"$/){$configValues{qMult}=$1;} 
	if(/^dgasURDir\s*=\s*\"(.*)\"$/){$configValues{dgasURDir}=$1;}
	if(/^dgasErrDir\s*=\s*\"(.*)\"$/){$configValues{dgasErrDir}=$1;}
	if(/^lockFileName\s*=\s*\"(.*)\"$/){$configValues{lockFileName}=$1;}
	if(/^mainPollInterval\s*=\s*\"(.*)\"$/){$configValues{mainPollInterval}=$1;}
	if(/^queuePollInterval\s*=\s*\"(.*)\"$/){$configValues{queuePollInterval}=$1;}
	if(/^lrmsType\s*=\s*\"(.*)\"$/){$configValues{lrmsType}=$1;}
	if(/^defaultVOUserHLRlist\s*=\s*\"(.*)\"$/){$configValues{defaultVOUserHLRlist}=$1;}
	if(/^defaultUserHLR\s*=\s*\"(.*)\"$/){$configValues{defaultUserHLR}=$1;}
	if(/^forceLocalOnly\s*=\s*\"(.*)\"$/){$configValues{forceLocalOnly}=$1;}
	if(/^gridUser\s*=\s*\"(.*)\"$/){$configValues{gridUser}=$1;}
	if(/^siteName\s*=\s*\"(.*)\"$/){$configValues{siteName}=$1;}
	if(/^localHostNameAsCEHostName\s*=\s*\"(.*)\"$/){$configValues{localHostNameAsCEHostName}=$1;}
    }
    close(FILE);

    # if we have VO-specific UserHLRs:
    if ($configValues{defaultVOUserHLRlist} ne "") {
	open(FILE, "$configValues{defaultVOUserHLRlist}") || &error("file not open : $configValues{defaultVOUserHLRlist}\n");

	print "".localtime().": Loading mapping for VO -> default User HLR:\n";
	while(<FILE>)
	{
	    next if (/^\s*\#+/);
	    if (/^\s*([^\s]*)\s*=\s*\"(.*)\"\s*$/) {
		$voDefUserHLRs{$1}=$2;
		print "VO '$1' has default User HLR: '$voDefUserHLRs{$1}'\n";
	    }
	}
	close(FILE);

	if ($configValues{defaultUserHLR} ne "") {
	    print "Generic default User HLR for all other VOs: '$configValues{defaultUserHLR}'\n";
	}
    }

}


sub checkQueue
{
    foreach (@queue)
    {
	my $file = $_;
	if($file =~ /($_[0]:\d*:\d*)/){ return 1;}   ## file in queue
    }
    
    return 0;
}      


sub pushToQueue 
{
    my $file= "$_[0]:0:0";
    unshift @queue, $file; 
    
    return 0;
}


# replaced, see below!
#sub parseFile
#{
#       my $file = $_[0];
#       my $queueName = "";
#	if ( $file =~ "pbs_" )
#	{
#		#This is a UR file for pbs LRMS
#		return &parseFile_PBS($file);
#		
#	}
#	if ( $file =~ "lsf_" )
#	{
#		#This is a UR file for lsf LRMS
#		return &parseFile_LSF($file);
#	}
#	#unknown file type
#	return 1;
#}

sub parseFile
{
    my $file = $_[0];

    # empty these from the previous processed record:
    $UR = "";
    %urGridInfo = ();
    %urAcctlogInfo = ();

    #print "Trying to open file: $dgas_dir/$_[0]\n";

    open(FILE, "< $dgas_dir/$_[0]") || return 1;

    print "\n".localtime().": Trying to parse UR file $dgas_dir/$_[0]\n";

    my $line;
    while ($line = <FILE>) {
	if ($line =~ /^ACCTLOG:(.*)$/ ) {
	    $UR = $1;
	    last; # stop here
	}
	if ($line =~ /^([^=]*)=(.*)$/) {
	    $urGridInfo{$1}=$2;
	}
    }

    close(FILE);

    if ($UR eq "") {
	print "Error: UR file has wrong format: $dgas_dir/$_[0] ... skipping!\n";
	return 2;
    }

    # try to determine the LRMS type.
    my $lrmsType = "";
    if (exists($urGridInfo{"LRMS_TYPE"}) && ($urGridInfo{"LRMS_TYPE"} ne "")) {
	# give priority to what is stated by Gianduia
	$lrmsType = $urGridInfo{"LRMS_TYPE"};
    }
    elsif ($file =~ /^([^_]*)_/) {
	# try to extract from file name
	$lrmsType = $1;
    }

    # verify that we can treat this LRMS type!
    if ($lrmsType ne "pbs" && $lrmsType ne "lsf") {
	print "Error: cannot determine LRMS type for UR file $dgas_dir/$_[0] ... skipping!\n";
	return 3;
    }
    else {
	#$urAcctlogInfo{jobmanagerType} = "blah";
	$urAcctlogInfo{lrmsType} = $lrmsType;
	print "LRMS type: $lrmsType\n";
    }

    &parseUR($lrmsType, $UR);

    return 0;
}


# not used anymore!
#sub parseFile_PBS 
#{
#    $UR   = "";
#    %hash = (
#	     server   =>  "server",
#	     queue    =>  "queue",
#	     lrmsType =>  "pbs",
#	     jobmanagerType =>  "blah",
#             globusResContact =>  "GLOBUS_RES_CONTACT",
#	     cput     =>  "cput",
#	     hlrLoc   =>  "HLR_LOCATION",
#	     jobId    =>  "WMS_JOBID",
#	     jobIdEdg =>  "EDG_WL_JOBID",
#	     user     =>  "PBS_O_LOGNAME",
#	     mem      =>  "mem",
#	     vmem     =>  "vmem", 
#	     walltime =>  "walltime",		   
#	     lrmsId   =>  "lrmsId",		   
#	     jobName  =>  "jobName",		  
#	     group    =>  "group", 
#	     ctime    =>  "ctime", 
#	     qtime    =>  "qtime", 
#	     etime    =>  "etime", 
#	     start    =>  "start", 
#	     end      =>  "end", 
#	     exitStatus      =>  "exitStatus", 
#	     execHost    =>  "execHost", 
#	     processors => "processors", 
#	     si2k => "GlueHostBenchmarkSI00", 
#	     sf2k => "GlueHostBenchmarkSF00", 
#	     tz => "timeZone", 
#	     );    
#    
#    open(FILE, "< $dgas_dir/$_[0]") || return 1;
#    
#    while(<FILE>)
#    {
#	if (/^$hash{jobId}=(.*)$/) #get GLITE_WMS_JOBID
#	{
#	#job coming from a gLite RB 
#		$hash{jobId}=$1;
#	}
#	if ( $hash{jobId} ne "WMS_JOBID" )
#	{
#                my $var="ACCTLOG:";
#                if (/^$var(.*)$/){$UR=$1;}#get the UR string
#	}
#	if (/^$hash{jobIdEdg}=(.*)$/)#get EDG_WL_JOBID
#	{
#	#job coming from a edg-like RB
#		$hash{jobIdEdg}=$1;
#	}
#	if ( $hash{jobIdEdg} ne "EDG_WL_JOBID" )
#	{
#		my $var="ACCTLOG:";
#		if (/^$var(.*)$/){$UR=$1;}#get the UR string
#	}
#	if (/^$hash{hlrLoc}=(.*)$/){$hash{hlrLoc}=$1};#get HLR_LOCATION
#	if (/^$hash{globusResContact}=(.*)$/){$hash{globusResContact}=$1};#get GLOBUS_RES_CONTACT
#	if (/^$hash{si2k}=(.*)$/){$hash{si2k}=$1};
#	if (/^$hash{sf2k}=(.*)$/){$hash{sf2k}=$1};
#	if (/^$hash{tz}=(.*)$/){$hash{tz}=$1};
#    }
#    close(FILE);
#    if ( $hash{jobId} ne "WMS_JOBID" )
#    {
#	#gLite	
#    	if ( $UR ne "" )
#    	{
#		print "parseUR_PBS\n";
# 		&parseUR_PBS($UR);
#    	}
#    }
#    if ( $hash{jobIdEdg} ne "EDG_WL_JOBID" )
#    {
#	#edg-like
#	if ( $UR ne "" )
#	{
#		&parseUR_PBS($UR);
#	}
#    }
#    return 0;
#}

# not used anymore:
#sub parseFile_LSF
#{
#	#FIXME
#	$UR   = "";
#	%hash = (
#		server   =>  "server",
#		queue    =>  "queue",
#	        lrmsType =>  "lsf",
#	     	jobmanagerType =>  "blah",
#		globusResContact =>  "GLOBUS_RES_CONTACT",
#	    	hlrLoc   =>  "HLR_LOCATION",
#	     	jobId    =>  "WMS_JOBID",
#	     	user     =>  "PBS_O_LOGNAME",
#	     	mem      =>  "mem",
#	     	vmem     =>  "vmem", 
#		cput     =>  "cput",
#	     	walltime =>  "walltime",		   
#	     	lrmsId   =>  "lrmsId",		  
#	        jobName  =>  "jobName",		  
#   	        group    =>  "group", 
#	        ctime    =>  "ctime", 
#	        qtime    =>  "qtime", 
#	        etime    =>  "etime", 
#	        start    =>  "start", 
#	        end      =>  "end", 
#	        exitStatus      =>  "exitStatus", 
#	        execHost    =>  "execHost", 
#		processors => "processors", 
#	        si2k => "GlueHostBenchmarkSI00", 
#	        sf2k => "GlueHostBenchmarkSF00", 
#	        tz => "timeZone", 
#	     );    
#	open(FILE, "< $dgas_dir/$_[0]") || return 1;
#	while(<FILE>)#FIXME NEW FILE FORMAT
#	{
#		if (/^$hash{hlrLoc}=(.*)$/){$hash{hlrLoc}=$1};#get HLR_LOCATION
#		if (/^$hash{jobId}=(.*)$/){$hash{jobId}=$1};#get EDG_WL_JOBID
#		if (/^$hash{globusResContact}=(.*)$/){$hash{globusResContact}=$1};#get GLOBUS_RES_CONTACT
#		if (/^$hash{si2k}=(.*)$/){$hash{si2k}=$1};
#		if (/^$hash{sf2k}=(.*)$/){$hash{sf2k}=$1};
#		if (/^$hash{tz}=(.*)$/){$hash{tz}=$1};
#		if (/^ACCTLOG:(.*)$/){$UR=$1};#get the UR string
#	}
#	close(FILE);
#	if ( $UR ne "" )
#	{
#		&parseUR_LSF($UR);
#	}
#	return 0;
#}


sub parseUR {
    if ($_[0] eq "pbs") {
	return &parseUR_pbs($_[1]);
    }
    elsif ($_[0] eq "lsf") {
	return &parseUR_lsf($_[1]);
    }
}



sub parseUR_pbs
{
	my $URString = $_[0];
	print "".localtime().": Got UR string:\n$URString\n";

	my @URArray = split ( ' ' , $URString );
	my @tmpArray = split ( ';', $URArray[1] );
	$_ = $tmpArray[3];
	if (/^user=(.*)$/){$urAcctlogInfo{user}=$1;};
	$urAcctlogInfo{lrmsId}=$tmpArray[2];
	$_ = $tmpArray[2];
	if (/^(\d*)\.(.*)$/){$urAcctlogInfo{server}=$2;};
	foreach ( @URArray )
	{
		if ( /^queue=(.*)$/){$urAcctlogInfo{queue}=$1;};
		if ( /^resources_used.cput=(.*)$/)
		{
			$_=$1;
			$_=~ /(\d\d):(\d\d):(\d\d)$/;
			$urAcctlogInfo{cput}= $3 + $2*60 + $1*3600;
		}
		if ( /^resources_used.walltime=(.*)$/)
		{
			$_=$1;
			$_=~ /(\d\d):(\d\d):(\d\d)$/;
			$urAcctlogInfo{walltime}= $3 + $2*60 + $1*3600;
		}
		if ( /^resources_used.vmem=(.*)$/)
		{
			$_=$1;
                        $_=~ /(\d*[M.k]b)$/;
			$urAcctlogInfo{vmem}     = $1;
		}
		if ( /^resources_used.mem=(.*)$/)
		{
			$_=$1;
			$_=~ /(\d*[M.k]b)$/;
			$urAcctlogInfo{mem}     = $1;
		}
		if ( /^Resource_List.neednodes=(\d*)$/)
		{	
			$urAcctlogInfo{processors} = $1;
			# attention! might also be list of hostnames,
			# in this case the number of hosts should be
			# counted!? What about SMP machines; is their
			# hostname listed N times or only once??
		}
		if ( /^group=(.*)$/)
		{	
			$urAcctlogInfo{group} = $1;
		}
		if ( /^jobname=(.*)$/)
		{	
			$urAcctlogInfo{jobName} = $1;
		}
		if ( /^ctime=(\d*)$/)
		{	
			$urAcctlogInfo{ctime} = $1;
		}
		if ( /^qtime=(\d*)$/)
		{	
			$urAcctlogInfo{qtime} = $1;
		}
		if ( /^etime=(\d*)$/)
		{	
			$urAcctlogInfo{etime} = $1;
		}
		if ( /^start=(\d*)$/)
		{	
			$urAcctlogInfo{start} = $1;
		}
		if ( /^end=(\d*)$/)
		{	
			$urAcctlogInfo{end} = $1;
		}
		if ( /^exec_host=(.*)$/)
		{	
			$urAcctlogInfo{execHost} = $1;
		}
		if ( /^Exit_status=(\d*)$/)
		{	
			$urAcctlogInfo{exitStatus} = $1;
		}
	}
	
}

sub parseUR_lsf
{
        my $URString = $_[0];
	print "".localtime()."Got UR string:\n$URString\n";

        my @new = quotewords(" ", 0, $URString);
        my $shift1 = $new[22];
        my $shift2 = $new[23+$shift1];
	#$urAcctlogInfo{server}=$new[16].".".$domainName;
	# attention: LSF log might already contain a complete hostname
	# (with domain name) ...
	# cross-check, instead of just adding the CE's domain name:
	$urAcctlogInfo{server}=$new[16];
	if ($urAcctlogInfo{server} =~ /^([^\.]*)\.(.*)$/)
	{
	    # if we found a domain name in the LSF log:
	    if ( ($2 ne $domainName) && ($domainName ne "") )
	    {
		# replace by CE's domain name if different:
		$urAcctlogInfo{server} = "$1.$domainName";
	    }
	}
	elsif ($domainName ne "")
	{
	    # if no domain name in LSF log: add the CE's domain name
	    $urAcctlogInfo{server} .= ".$domainName";
	}

        $urAcctlogInfo{queue}=$new[12];
        $urAcctlogInfo{user}=$new[11];
	# we now know the user's UNIX login, try to find out the group, that
	# isn't available in the LSF log file:
	my $groupOutput = `groups $urAcctlogInfo{user}`;
	if ($groupOutput =~ /^$urAcctlogInfo{user} : ([^\s]+).*$/) {
	    $urAcctlogInfo{group} = $1;
	}

        $urAcctlogInfo{lrmsId}=$new[3];
        $urAcctlogInfo{processors}=$new[6];
	if ($new[10] == 0)
	{   # happens if job was cancelled before executing?
	    $urAcctlogInfo{walltime}=0;
	}
	else
	{
	    $urAcctlogInfo{walltime}=$new[2]-$new[10];
	}
	if ($new[28+$shift2] == -1)
	{   # indicates that the value is not available!
	    $new[28+$shift2]=0;
	}
	if ($new[29+$shift2] == -1)
	{   # indicates that the value is not available!
	    $new[29+$shift2]=0;
	}
        $urAcctlogInfo{cput}=int($new[28+$shift2])+int($new[29+$shift2]);
        $urAcctlogInfo{mem}=$new[54+$shift2]."k";
        $urAcctlogInfo{vmem}=$new[55+$shift2]."k";
        $urAcctlogInfo{start}=$new[10];
        $urAcctlogInfo{end}=$new[2];
        $urAcctlogInfo{ctime}=$new[7];
        $urAcctlogInfo{jobName}=$new[26+$shift2];
        $urAcctlogInfo{exitStatus}=$new[49+$shift2];
}


sub callAtmClient 
{
    my $now_string = localtime();
    if ( $keepGoing == 0 )
    {
	print "$now_string: Got termination signal..\n";
	return -2;
    }
    my $cmd = "";
    my $status = -1;
    my $resGridId= "";

    my $forceThisJobLocal = "no";
    if (exists($urGridInfo{JOB_TYPE}) && $urGridInfo{JOB_TYPE} eq "local")
    {
	print "$now_string: JOB_TYPE=local found in UR file. Storing UR on Resource HLR _only_!\n";
	$forceThisJobLocal = "yes";
    }

    my $hasProxy = 0; # new sensors don't furnish the proxy file
    if (-r "$dgas_dir/$_[0].proxy") {
	$hasProxy = 1;
    }

    if ($configValues{localHostNameAsCEHostName} eq "yes") {
	$urAcctlogInfo{server} = $hostName.".".$domainName;
    }

    # CE ID:
    if (exists($urGridInfo{CE_ID}) && $urGridInfo{CE_ID} ne "") {
	$resGridId = $urGridInfo{"CE_ID"};
	print "$now_string: Got CE_ID from UR file: $resGridId\n";
    }
    elsif (exists($urGridInfo{GLOBUS_RES_CONTACT})
	   && $urGridInfo{GLOBUS_RES_CONTACT} ne ""
	   && $urAcctlogInfo{queue} ne "") {
	$resGridId = $urGridInfo{GLOBUS_RES_CONTACT};
	print "$now_string: Got GLOBUS_RES_CONTACT from UR file: $resGridId\n";
    }
    elsif ($forceThisJobLocal
	   && exists($urAcctlogInfo{server}) && $urAcctlogInfo{server} ne ""
	   && exists($urAcctlogInfo{queue}) && $urAcctlogInfo{queue} ne "") {
	$resGridId = "$urAcctlogInfo{server}:$urAcctlogInfo{queue}";
	print "$now_string: Constructed resource ID for local job: $resGridId\n";
    }
    else {
	print "Error: cannot determine CE/resource ID! Skipping job!\n";
	return -3;
    }

    # does the CE ID finish with the queue name???
    if ($resGridId !~ /[-:]$urAcctlogInfo{queue}\W*$/) {
	$resGridId = $resGridId."-".$urAcctlogInfo{queue};
	print "Added queue name: CE_ID = $resGridId\n";
    }

    #print "$now_string: Found CE ID: $resGridId\n";

    # User DN:
    my $userDN = "";
    if (exists($urGridInfo{USER_DN}) && $urGridInfo{USER_DN} ne "") {
	$userDN = $urGridInfo{USER_DN};
	print "$now_string: Got USER_DN from UR file: $userDN\n";
    }
    elsif ($hasProxy) {
	&checkProxy("$dgas_dir/$_[0].proxy", $userDN);
	if ($userDN ne "") {
	    print "$now_string: Got USER_DN from proxy file: $userDN\n";
	}
	else {
	    print "$now_string: Couldn't determine USER_DN.\n";
	}
    }
    else {
	print "$now_string: Couldn't determine USER_DN.\n";
    }

    # FQAN:
    my $fqan = "";
    if (exists($urGridInfo{USER_FQAN}) && $urGridInfo{USER_FQAN} ne "") {
	$fqan = $urGridInfo{USER_FQAN};
	print "$now_string: Got USER_FQAN from UR file: $fqan\n";
    }
    elsif ($hasProxy && (&findUserFqan("$dgas_dir/$_[0].proxy",$fqan) == 0)) {
	print "$now_string: Got USER_FQAN from proxy file: $fqan\n";
    }
    else {
	print "$now_string: Couldn't determine USER_FQAN.\n";
    }

    # VO:
    my $userVo = "";
    if ( &determineUserVO($fqan, $urAcctlogInfo{user}, $userVo) == 0 ) {
	print "$now_string: Determined user VO: $userVo\n";
    }
    
    # User HLR location: 
    if (!exists($urGridInfo{HLR_LOCATION}) || $urGridInfo{HLR_LOCATION} eq "" )
    {
	# no HLRLocation specified in the JDL expression, try defaults:
	# first see if there is a default User HLR for the VO:
	if(exists($voDefUserHLRs{$userVo})) {
	    print "$now_string: Using default HLR Location for VO $userVo: $voDefUserHLRs{$userVo}\n";
	    $urGridInfo{HLR_LOCATION} = $voDefUserHLRs{$userVo};
	}
	# if no VO-specific User HLR try the general default value:
	elsif ($defaultUserHLR ne "" )
        {
		print "$now_string: Using generic default HLR Location for users: $defaultUserHLR\n";
                $urGridInfo{HLR_LOCATION} = $defaultUserHLR;
        }
	else
	{
		print "$now_string: User didn't specify HLRLocation in the JDL and defaultUserHLR is empty. Storing UR on Resource HLR _only_!\n";
		$forceThisJobLocal = "yes";
	}
    }

    # grid Job ID:
    my $gridJobId = "";
    if (exists($urGridInfo{GRID_JOBID}) && $urGridInfo{GRID_JOBID} ne "") {
	$gridJobId = $urGridInfo{GRID_JOBID};
	print "$now_string: Got GRID_JOBID from UR file: $gridJobId\n";
    }
    elsif (exists($urGridInfo{WMS_JOBID}) && $urGridInfo{WMS_JOBID} ne "") {
	$gridJobId = $urGridInfo{WMS_JOBID};
	print "$now_string: Got WMS_JOBID from UR file: $gridJobId\n";
    }
    elsif (exists($urGridInfo{EDG_WL_JOBID}) && $urGridInfo{EDG_WL_JOBID} ne "") {
	$gridJobId = $urGridInfo{EDG_WL_JOBID};
	print "$now_string: Got EDG_WL_JOBID from UR file: $gridJobId\n";
    }
    elsif (exists($urAcctlogInfo{server}) && $urAcctlogInfo{server} ne ""
	   && exists($urAcctlogInfo{lrmsId}) && $urAcctlogInfo{lrmsId} ne ""
	   && exists($urAcctlogInfo{start}) && $urAcctlogInfo{start} ne "") {

	$gridJobId = $urAcctlogInfo{server}.":".$urAcctlogInfo{lrmsId}."_".$urAcctlogInfo{start};

	print "$now_string: No grid job ID. Constructed a  unique ID: $gridJobId\n";
    }
    else {
	print "Error: cannot retrieve or construct a unique job ID! Skipping job\n";
	return -4;
    }

    # building command:

    $cmd = "/opt/glite/libexec/glite_dgas_atmClient";

    if ($gridJobId ne "") {
        $cmd .= " --jobid \"$gridJobId\"";
    }
    if ($resGridId ne "") {
	$cmd .= " --resgridid \"$resGridId\"";
    }
    if (exists($urGridInfo{HLR_LOCATION}) && $urGridInfo{HLR_LOCATION} ne "") {
	$cmd .= " --remhlrid \"$urGridInfo{HLR_LOCATION}\"";
    }
    if (exists($urAcctlogInfo{cput})) {
	$cmd .= " \"CPU_TIME=$urAcctlogInfo{cput}\"";
    }
    if (exists($urAcctlogInfo{walltime})) {
	$cmd .= " \"WALL_TIME=$urAcctlogInfo{walltime}\"";
    }
    if (exists($urAcctlogInfo{mem})) {
	$cmd .= " \"PMEM=$urAcctlogInfo{mem}\"";
    }
    if (exists($urAcctlogInfo{vmem})) {
	$cmd .= " \"VMEM=$urAcctlogInfo{vmem}\"";
    }
    if (exists($urAcctlogInfo{queue})) {
	$cmd .= " \"QUEUE=$urAcctlogInfo{queue}\"";
    }
    if (exists($urAcctlogInfo{user})) {
	$cmd .= " \"USER=$urAcctlogInfo{user}\"";
    }
    if (exists($urAcctlogInfo{lrmsId})) {
	$cmd .= " \"LRMSID=$urAcctlogInfo{lrmsId}\"";
    }
    if (exists($urAcctlogInfo{processors})) {
	$cmd .= " \"PROCESSORS=$urAcctlogInfo{processors}\"";
    }
    $cmd .= " \"URCREATION=$now_string\"";
    if (exists($urAcctlogInfo{group})) {
	$cmd .= " \"group=$urAcctlogInfo{group}\"";
    }
    if (exists($urAcctlogInfo{jobName})) {
	$cmd .= " \"jobName=$urAcctlogInfo{jobName}\"";
    }
    if (exists($urAcctlogInfo{start})) {
	$cmd .= " \"start=$urAcctlogInfo{start}\"";
    }
    if (exists($urAcctlogInfo{end})) {
	$cmd .= " \"end=$urAcctlogInfo{end}\"";
    }
    if (exists($urAcctlogInfo{ctime})) {
	$cmd .= " \"ctime=$urAcctlogInfo{ctime}\"";
    }
    if (exists($urAcctlogInfo{qtime})) {
	$cmd .= " \"qtime=$urAcctlogInfo{qtime}\"";
    }
    if (exists($urAcctlogInfo{etime})) {
	$cmd .= " \"etime=$urAcctlogInfo{etime}\"";
    }
    if (exists($urAcctlogInfo{exitStatus})) {
	$cmd .= " \"exitStatus=$urAcctlogInfo{exitStatus}\"";
    }

    if (exists($urAcctlogInfo{execHost})) {
	$cmd .= " \"execHost=$urAcctlogInfo{execHost}\"";
    }

    if (exists($urGridInfo{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"})
	&& $urGridInfo{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"} ne "") {
	my $si2k = $urGridInfo{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"};
	$cmd .= " \"si2k=$si2k\"";
	print "$now_string: Got 'GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU' from UR file: $si2k\n";
	print "(note: Taking mean SI00, ignoring GlueHostBenchmarkSI00)\n";
    }
    elsif (exists($urGridInfo{GlueHostBenchmarkSI00})
	&& $urGridInfo{GlueHostBenchmarkSI00} ne "") {
	$cmd .= " \"si2k=$urGridInfo{GlueHostBenchmarkSI00}\"";
	print "$now_string: Got 'GlueHostBenchmarkSI00' from UR file: $urGridInfo{GlueHostBenchmarkSI00}\n";
    }

    if (exists($urGridInfo{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"})
	&& $urGridInfo{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"} ne "") {
	my $sf2k = $urGridInfo{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"};
	$cmd .= " \"sf2k=$sf2k\"";
	print "$now_string: Got 'GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU' from UR file: $sf2k\n";
	print "(note: Taking mean SF00, ignoring GlueHostBenchmarkSF00)\n";
    }
    elsif (exists($urGridInfo{GlueHostBenchmarkSF00})
	&& $urGridInfo{GlueHostBenchmarkSF00} ne "") {
	$cmd .= " \"sf2k=$urGridInfo{GlueHostBenchmarkSF00}\"";
	print "$now_string: Got 'GlueHostBenchmarkSF00' from UR file: $urGridInfo{GlueHostBenchmarkSF00}\n";
    }

    if (exists($urGridInfo{timeZone})) {
	$cmd .= " \"tz=$urGridInfo{timeZone}\"";
    }
    if ($fqan ne "") {
        $cmd .= " \"fqan=$fqan\"";
    }
    if ($userVo ne "") {
        $cmd .= " \"userVo=$userVo\"";
    }
    if ( $configValues{siteName} ne "" )
    { 
            $cmd .= " \"SiteName=$configValues{siteName}\"";
    }


    # Is this a strictly local job? Let the HLR server know!
    if (exists($urGridInfo{JOB_TYPE}) && $urGridInfo{JOB_TYPE} eq "local")
    {
	$cmd .= " \"accountingProcedure=outOfBand\"";
    }


    # decide which UR routing and which local user to use:

    if ( ( $forceLocalOnly eq "yes" ) || ($forceThisJobLocal eq "yes") ) {
	# local job, only Resource HLR (route 3):
	print "$now_string: forceLocalOnly is set to \"yes\" in conf file or the job is specified as local in the UR. Therefore the UR is sent to the Resource(Site) HLR _only_.\n";
	if ($userDN ne "") {
	    $cmd .= " --usrcert \"$userDN\"";
	}
	$cmd .= " -3";
    }
    elsif ( $forceLocalFirst eq "yes" ) {
	# route 2 (Resource HLR -> User HLR) because requested:
	print "$now_string: forceLocalFirst is set to \"yes\". Therefore, the UR is sent to the Resource(Site) HLR first.\n";
	if ($userDN ne "") {
	    $cmd .= " --usrcert \"$userDN\"";
	}
	$cmd .= " -2";
    }
    elsif ( !$hasProxy
	    || (&checkProxy("$dgas_dir/$_[0].proxy", $userDN) != 0) ) {
	# route 2 (Resource HLR -> User HLR) because no proxy:
	print "$now_string: proxy file for job $urAcctlogInfo{lrmsId} with gridJobId $gridJobId is not present or invalid. Trying host certificate (Resource HLR first).\n";
	if ($userDN ne "") {
	    $cmd .= " --usrcert \"$userDN\"";
	}
	$cmd .= " -2";
    }
    else {
	# route 1 (User HLR -> Resource HLR). Need to use user proxy:
	print "$now_string: Trying to send UR to User HLR first, using proxy with userDN = $userDN\n";
	my $dgasUser = "";
	if ( $configValues{gridUser} eq "" ) {
	    $dgasUser = $urAcctlogInfo{user};
	}
	else {
	    $dgasUser = $configValues{gridUser};
	}
	 
	print "$now_string: Changing ownership of $_[0].proxy, to user: $dgasUser\n";
	`chown $dgasUser:root $dgas_dir/$_[0].proxy`;
	`chmod 600 $dgas_dir/$_[0].proxy`;

	$ENV{X509_USER_PROXY} = "$dgas_dir/$_[0].proxy";
	$cmd = "su $dgasUser -c '".$cmd."'";
    }

    # running the command: 

    print "$now_string: Running: $cmd\n";
    $status = system("$cmd");
    if ($? == -1) {
	print "failed to execute: $!\n";
    }
    elsif ($? & 127) {
	$status = ($? & 127);
    }
    else {
	$status = $? >> 8;
    }
    print "EXITSTATUS=$status\n";

    if (exists($ENV{X509_USER_PROXY})) {
	delete $ENV{X509_USER_PROXY};
    }

    if ( $status == 65 || ($status>= 70 && $status<=73) ) {
	print "The record is already present on the HLR. Removing ...\n";
	return 0;
    }

    return $status;
}


sub checkProxy() {
    my $proxyFile = $_[0];
    my $userDN = $_[1];
#   my $user = $_[1];

    my $openssl = `which openssl`;
    if ( -e $openssl ) {
	#print "Found $openssl\n";	
    }
    else {
	$openssl = "/usr/bin/openssl";
	#print "Trying with $openssl\n";
    }
    my $exitStatus = 0;
    if ( -e $openssl ) {
	my $opensslCmd = "$openssl x509 -in $proxyFile -noout -checkend 20";
	$exitStatus = system($opensslCmd);
    }
    else {
	print "Warning: couldn't find $openssl command\n";
    }
    $_ = `$openssl x509 -in $proxyFile -noout -subject`;
    s/(\/CN=(limited |)proxy)//g;
    if ( /^subject= (.*)$/) {
	my $buff = $1;
	$buff =~ s/emailAddress/Email/g;
	#print "Substituted email Address field: $buff\n"; 
	$_[1] = $buff;
    }
    #print "Openssl found user DN:$_[1]\n";
    return $exitStatus;
}


sub checkDgasUser()
{
    my $ulogin = $_[0];
    if ($ulogin eq "") {
	return -1;  # use pool accounts
    }

    if ( ($ulogin eq "root") || (system("su $ulogin -c 'true'") != 0) ) {
	return 1; # cannot use the user, use root with route 2!
    }

    return 0; # ok
}


sub findVomsProxyInfo()
{
        my $buff = `which voms-proxy-info`;
        $buff =~ s/\n//g;
        if ( -e $buff )
        {
                $_[0] = $buff;
                return 0;
        }
        else
        {
                $buff = "/opt/glite/bin/voms-proxy-info";
                if ( -e $buff )
                {
                        $_[0] = $buff;
                        return 0;
                }
        }
        return 1;
}


sub findUserFqan()
{
        if ( -e $vomsProxyInfo )
        {
                my $fqan = `$vomsProxyInfo -file $_[0] -fqan`;
                $fqan =~ s/\n/;/g;
		if ($fqan =~ /^(.*);$/)
		{
			$fqan = $1;
		}
                $_[1] = $fqan;
                return 0;
        }
        return 1;
}



sub determineUserVO()
{

    my $retVal = 1; # not found

    my $uFqan = $_[0];
    my $uid = $_[1];

    my @fqanParts = split(/\//, $uFqan);

    if (scalar(@fqanParts) > 1 && $fqanParts[1] ne "")
    {
	if ($fqanParts[1] =~ /^VO=(.*)$/)
	{
	    $fqanParts[1] = $1;
	}
	$_[2] = $fqanParts[1];
	$retVal = 0;
    }
    elsif ($uid =~ /^(.*)sgm$/ || $uid =~ /^(.*)prd$/ || $uid =~ /^(.*)\d{3}$/)
    {
	$_[2] = $1;
	$retVal = 0;	
    }

    return $retVal;
}


sub setFile()
{
    my @set = @_;   ## set[0] = fileName, set[1] = multiplicity, set[2] = depth
    
    if ($set[2] >= $qDept) 
    {
	$set[1]++;
	$set[2] = 0;
    }
    else {$set[2]++;}
    
    my $point = ":";
    my $updated = $set[0].$point.$set[1].$point.$set[2];
    
    return $updated;
}



sub delFile {   #receive as argument the filename to delete

    my $cmd = "rm -f $dgas_dir/$_[0]";
    print "$cmd \n";
    my $status = system($cmd);
    if (-f "$dgas_dir/$_[0].proxy") {
	$cmd .=".proxy";
	print "$cmd \n";
	$status =system($cmd);
    }
}



#sub printOut {
#
#    print "queue: $hash{queue} \n";      
#    print "server: $hash{server} \n";      
#    print "HLR_LOCATION: $hash{hlrLoc} \n";      
#    print "EDG_WL_JOBID: $hash{jobId} \n";      
#    print "cput: $hash{cput} \n";
#    print "mem: $hash{mem} \n";
#    print "vmem: $hash{vmem} \n";
#    print "walltime: $hash{walltime} \n";  
#    print "------------------------------------ \n";
#}



sub timeOut
{
    my $stop = time();
    my $delta = $stop -$start;
    #print "delta = $delta \n" ;   
    
    if($delta > $T2)
    {
	$start = time();
	return 0;
    }
    
    return 1;
}


sub createErrDir
{
    my $cmd = "mkdir $dgas_err_dir";
    print "$cmd \n";
    my $status = system($cmd);

    return $status;
}



sub moveInErrDir
{
    my $cmd = "mv $dgas_dir/$_[0] $dgas_err_dir";
    my $status = system($cmd);
    print "$cmd \n";
    if (-f "$dgas_dir/$_[0].proxy") {
	$cmd = "mv $dgas_dir/$_[0].proxy $dgas_err_dir";
	$status = system($cmd);
    }
    print "$cmd \n";

    return $status;
}



##-------> lock  subroutines <---------##

sub putLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") && return 1;
    close(IN);

    open(OUT, "> $lockName") || return 2;
    print OUT  $$;    ## writes pid   
    close(OUT);
    return 0;
}

sub delLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") || return 1;
    close(IN);
    my $status = system("rm -f $lockName");
    return $status;
}


##-------> sig handlers subroutines <---------##

sub sigHUP_handler {
        print "got SIGHUP\n";
	$keepGoing = 1;
	goto START;
}

sub sigINT_handler {
        print "got SIGINT\n";
        $keepGoing = 0;
}



## ----- print error and quit ----- ##

sub error {
    if (scalar(@_) > 0) {
	print "$_[0]";
    }
    exit(1);
}
